﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.DesktopEdition;

namespace SnapToGridMethods
{

    public partial class Form1 : Form
    {

        Proj4Projection proj = new Proj4Projection();
        private SnapToLayerEditInteractiveOverlay snapToLayerEditInteractiveOverlay = null;
        private double tolerance = 500;
        private bool inDrawMode = false;
        private bool inDrawComplexMode = false;

        public Form1()
        {
            InitializeComponent();

            winformsMap1.MapUnit = GeographyUnit.Meter;

            winformsMap1.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;

            proj.InternalProjectionParametersString = ManagedProj4Projection.GetWgs84ParametersString();
            proj.ExternalProjectionParametersString = ManagedProj4Projection.GetSphericalMercatorParametersString();

            GraticuleAdornmentLayer graticuleAdornmentLayer = new GraticuleAdornmentLayer();
            graticuleAdornmentLayer.GraticuleLineStyle.OuterPen.Color = GeoColor.FromArgb(125, GeoColor.StandardColors.White);
            graticuleAdornmentLayer.GraticuleTextFont = new GeoFont("Arial", 8, DrawingFontStyles.Bold);
            graticuleAdornmentLayer.Projection = proj;

            winformsMap1.AdornmentOverlay.Layers.Add("graticule", graticuleAdornmentLayer);

            winformsMap1.CurrentExtent = new RectangleShape(500000, 3100000, 600000, 3000000);

            snapToLayerEditInteractiveOverlay = CreateSnapToGrid(winformsMap1, 2000, 2000, proj, tolerance);

            winformsMap1.Refresh();
        }

        public static SnapToLayerEditInteractiveOverlay CreateSnapToGrid(WinformsMap winformsMap, Double inline,
            Double crossLine, Proj4Projection proj4Projection, double snapTolerance)
        {

            RectangleShape boundaryBox = winformsMap.CurrentExtent;
            boundaryBox.ScaleDown(80);
            PointShape startCorner = boundaryBox.UpperLeftPoint;

            //inMemoryFeatureLayer used to be snapped to.
            InMemoryFeatureLayer snappingGridLayer = new InMemoryFeatureLayer();
            snappingGridLayer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle = PointStyles.CreateSimplePointStyle(PointSymbolType.Cross, GeoColor.StandardColors.Black, GeoColor.StandardColors.Black, 1, 4);
            snappingGridLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            double boxWidth = boundaryBox.Width;
            double boxHeight = boundaryBox.Height;

            for (double i = 0; i <= boxHeight; i = i + crossLine)
            {
                for (double j = 0; j <= boxWidth; j = j + inline)
                {
                    PointShape newPoint = new PointShape(startCorner.X, startCorner.Y);
                    // Now add j to the Y value.
                    newPoint.X += j;
                    newPoint.Y -= i;
                    snappingGridLayer.InternalFeatures.Add(new Feature(newPoint));
                }
            }

            // SnapToLayerEditInteractiveOverlay to snap dragged control point to nearest point of layer if within tolerance.
            SnapToLayerEditInteractiveOverlay snapToLayerEditInteractiveOverlay = new SnapToLayerEditInteractiveOverlay();

            snapToLayerEditInteractiveOverlay.EditShapesLayer.InternalFeatures.Clear();

            LayerOverlay snappingGridOverlay = new LayerOverlay();
            snappingGridOverlay.Layers.Add("SnappingGrid", snappingGridLayer);
            snappingGridOverlay.Name = "SnappingGridOverlay";
            winformsMap.Overlays.Add("SnappingGridOverlay", snappingGridOverlay);

            //Sets the PointStyle for the non dragged control points.
            snapToLayerEditInteractiveOverlay.ControlPointStyle = new PointStyle(PointSymbolType.Circle, new GeoSolidBrush(GeoColor.StandardColors.PaleGoldenrod), new GeoPen(GeoColor.StandardColors.Black), 8);
            //Sets the PointStyle for the dragged control points.
            snapToLayerEditInteractiveOverlay.DraggedControlPointStyle = new PointStyle(PointSymbolType.Circle, new GeoSolidBrush(GeoColor.StandardColors.Red), new GeoPen(GeoColor.StandardColors.Orange, 2), 8);

            snapToLayerEditInteractiveOverlay.ToSnapInMemoryFeatureLayer = snappingGridLayer;

            //Example using World coordinates for tolerance.
            snapToLayerEditInteractiveOverlay.ToleranceType = ToleranceCoordinates.World;
            snapToLayerEditInteractiveOverlay.Tolerance = (float)snapTolerance;
            snapToLayerEditInteractiveOverlay.ToleranceUnit = DistanceUnit.Meter;

            snapToLayerEditInteractiveOverlay.CalculateAllControlPoints();

            winformsMap.EditOverlay = snapToLayerEditInteractiveOverlay;
            winformsMap.Refresh();

            return snapToLayerEditInteractiveOverlay;
        }

        private void DrawingTool_ItemClick(object sender, EventArgs e)
        {
            
            // Initialize a new track layer because ...
            // Clearing the overlays isn't enough, we have to basically recreate the layer
            TrackInteractiveOverlay track = new TrackInteractiveOverlay();

            // The Track Layer Style
            Color baseColor = Color.Blue;
            GeoColor geoBaseColor = new GeoColor(175, baseColor.R, baseColor.G, baseColor.B);

            track.TrackShapeLayer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle =
                AreaStyles.CreateSimpleAreaStyle(geoBaseColor, geoBaseColor, 2);
            track.TrackShapeLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            // Enable an event to allow us to snap to the nearest snap grid point if it exists
            //track.TrackEnding += new EventHandler<TrackEndingTrackInteractiveOverlayEventArgs>(TrackEnding);
            //track.TrackEnded += new EventHandler<TrackEndedTrackInteractiveOverlayEventArgs>(TrackEnded);
            //track.TrackStarted += new EventHandler<TrackStartedTrackInteractiveOverlayEventArgs>(TrackStarted);
            //track.TrackStarting += new EventHandler<TrackStartingTrackInteractiveOverlayEventArgs>(TrackStarting);
            //track.VertexAdding += new EventHandler<VertexAddingTrackInteractiveOverlayEventArgs>(VertexAdding);
            //track.VertexAdded += new EventHandler<VertexAddedTrackInteractiveOverlayEventArgs>(VertexAdded);

            winformsMap1.TrackOverlay = track;

            string buttonName = ((ToolStripMenuItem)sender).Name;

            switch (buttonName)
            {
                case "tspPoint":
                    // THIS WORKS!
                    track.TrackEnding += new EventHandler<TrackEndingTrackInteractiveOverlayEventArgs>(TrackEnding);
                    winformsMap1.TrackOverlay.TrackMode = TrackMode.Point;
                    inDrawMode = true;
                    break;
                case "tspLine":
                    // THIS WORKS!
                    track.TrackEnding += new EventHandler<TrackEndingTrackInteractiveOverlayEventArgs>(TrackEnding);
                    track.VertexAdding += new EventHandler<VertexAddingTrackInteractiveOverlayEventArgs>(VertexAdding);
                    winformsMap1.TrackOverlay.TrackMode = TrackMode.StraightLine;
                    inDrawMode = true;
                    break;
                case "tspRectangle":
                    // ONLY THE FIRST POINT WILL SNAP
                    track.TrackEnding += new EventHandler<TrackEndingTrackInteractiveOverlayEventArgs>(TrackEnding);
                    track.VertexAdding += new EventHandler<VertexAddingTrackInteractiveOverlayEventArgs>(VertexAdding);
                    winformsMap1.TrackOverlay.TrackMode = TrackMode.Rectangle;
                    inDrawMode = true;
                    break;
                case "tspPolygon":
                    // ALL POINTS EXCEPT THE LAST POINT SNAP
                    track.TrackEnding += new EventHandler<TrackEndingTrackInteractiveOverlayEventArgs>(TrackEnding);
                    track.VertexAdding += new EventHandler<VertexAddingTrackInteractiveOverlayEventArgs>(VertexAdding);
                    winformsMap1.TrackOverlay.TrackMode = TrackMode.Polygon;
                    inDrawComplexMode = true;
                    break;
                case "tspMultiLine":
                    // ONLY THE START AND END POINTS WILL SNAP EVEN THOUGH THE ADDED VERTEX 
                    // ITEMS APPEAR TO SNAP AT FIRST
                    track.TrackEnding += new EventHandler<TrackEndingTrackInteractiveOverlayEventArgs>(TrackEnding);
                    track.VertexAdding += new EventHandler<VertexAddingTrackInteractiveOverlayEventArgs>(VertexAdding);
                    winformsMap1.TrackOverlay.TrackMode = TrackMode.Line;
                    inDrawComplexMode = true;
                    break;
            }

        }

        private void TrackStarted(object sender, TrackStartedTrackInteractiveOverlayEventArgs e)
        {
            Debug.Print("TrackStarted");
            PointShape closestPoint = FindNearestSnappingPoint(new PointShape(e.StartedVertex));
            if (closestPoint == null) return;
            e.StartedVertex = new Vertex(closestPoint);
        }

        private void TrackStarting(object sender, TrackStartingTrackInteractiveOverlayEventArgs e)
        {
            Debug.Print("TrackStarting");
            PointShape closestPoint = FindNearestSnappingPoint(new PointShape(e.StartingVertex));
            if (closestPoint == null) return;
            e.StartingVertex = new Vertex(closestPoint);
        }

        private void TrackEnding(object sender, TrackEndingTrackInteractiveOverlayEventArgs e)
        {
            Debug.Print("TrackEnding");
            WellKnownType type = e.TrackShape.GetWellKnownType();
            PointShape closestPoint = null;

            switch (type)
            {
                case WellKnownType.Line:
                    LineShape line = new LineShape(e.TrackShape.GetWellKnownBinary());
                    int count = line.Vertices.Count() - 1;
                    closestPoint = FindNearestSnappingPoint(new PointShape(line.Vertices[count]));
                    if (closestPoint == null) return;
                    line.Vertices[count] = new Vertex(closestPoint);
                    e.TrackShape = line;
                    break;
                case WellKnownType.Point:
                    PointShape currentPoint = new PointShape(e.TrackShape.GetWellKnownBinary());
                    closestPoint = FindNearestSnappingPoint(currentPoint);
                    if (closestPoint == null) return;
                    e.TrackShape = closestPoint;
                    break;
                case WellKnownType.Polygon:
                    PolygonShape poly = new PolygonShape(e.TrackShape.GetWellKnownBinary());
                    int pointCount = poly.OuterRing.Vertices.Count() - 1;
                    closestPoint = FindNearestSnappingPoint(new PointShape(poly.OuterRing.Vertices[pointCount]));
                    if (closestPoint == null) return;
                    poly.OuterRing.Vertices[pointCount] = new Vertex(closestPoint);
                    e.TrackShape = poly;
                    break;
                default:
                    break;
            }

        }

        private void TrackEnded(object sender, TrackEndedTrackInteractiveOverlayEventArgs e)
        {
            Debug.Print("TrackEnded");
            WellKnownType type = e.TrackShape.GetWellKnownType();
            PointShape closestPoint = null;

            switch (type)
            {
                case WellKnownType.Line:
                    LineShape line = new LineShape(e.TrackShape.GetWellKnownBinary());
                    int count = line.Vertices.Count() - 1;
                    if (count > 2) return;
                    closestPoint = FindNearestSnappingPoint(new PointShape(line.Vertices[count]));
                    if (closestPoint == null) return;
                    line.Vertices[count] = new Vertex(closestPoint);
                    e.TrackShape = line;
                    break;
                case WellKnownType.Point:
                    PointShape currentPoint = new PointShape(e.TrackShape.GetWellKnownBinary());
                    closestPoint = FindNearestSnappingPoint(currentPoint);
                    if (closestPoint == null) return;
                    e.TrackShape = closestPoint;
                    break;
                case WellKnownType.Polygon:
                    PolygonShape poly = new PolygonShape(e.TrackShape.GetWellKnownBinary());
                    int pointCount = poly.OuterRing.Vertices.Count() - 1;
                    closestPoint = FindNearestSnappingPoint(new PointShape(poly.OuterRing.Vertices[pointCount]));
                    if (closestPoint == null) return;
                    poly.OuterRing.Vertices[pointCount] = new Vertex(closestPoint);
                    e.TrackShape = poly;
                    break;
                default:
                    break;
            }

        }

        private void VertexAdding(object sender, VertexAddingTrackInteractiveOverlayEventArgs e)
        {
            Debug.Print("VertexAdding");
            WellKnownType type = e.AffectedFeature.GetWellKnownType();
            //if (type == WellKnownType.Point) return;

            PointShape closestPoint = FindNearestSnappingPoint(new PointShape(e.AddingVertex));
            if (closestPoint == null) return;
            e.AddingVertex = new Vertex(closestPoint);
        }

        private void VertexAdded(object sender, VertexAddedTrackInteractiveOverlayEventArgs e)
        {
            Debug.Print("VertexAdded");
            PointShape closestPoint = FindNearestSnappingPoint(new PointShape(e.AddedVertex));
            if (closestPoint == null) return;
            e.AddedVertex = new Vertex(closestPoint);
        }

        private PointShape FindNearestSnappingPoint(PointShape targetPointShape)
        {
            InMemoryFeatureLayer imfl = snapToLayerEditInteractiveOverlay.ToSnapInMemoryFeatureLayer;

            imfl.Open();
            Collection<Feature> toSnapInMemoryFeatures = imfl.FeatureSource.GetFeaturesNearestTo(targetPointShape, GeographyUnit.Meter, 1, ReturningColumnsType.AllColumns);
            imfl.Close();

            if (toSnapInMemoryFeatures.Count == 1)
            {
                BaseShape baseShape = toSnapInMemoryFeatures[0].GetShape();

                string baseShapeType = baseShape.GetWellKnownType().ToString();

                if (baseShapeType == "Multipoint")
                {
                    MultipointShape multiPointShape = (MultipointShape)toSnapInMemoryFeatures[0].GetShape();

                    foreach (PointShape point in multiPointShape.Points)
                    {
                        PointShape toSnapPointShape = point;
                        double Distance = toSnapPointShape.GetDistanceTo(targetPointShape, GeographyUnit.Meter, DistanceUnit.Meter);

                        if (Distance <= tolerance)
                        {
                            return new PointShape(toSnapPointShape.X, toSnapPointShape.Y);
                        }
                    }
                }
                else if (baseShapeType == "Point")
                {
                    PointShape toSnapPointShape = (PointShape)baseShape;
                    double Distance = toSnapPointShape.GetDistanceTo(targetPointShape, GeographyUnit.Meter, DistanceUnit.Meter);

                    if (Distance <= tolerance)
                    {
                        return new PointShape(toSnapPointShape.X, toSnapPointShape.Y);
                    }
                }
            }
            return null;

        }

        private void winformsMap1_DoubleClick(object sender, EventArgs e)
        {
            if (inDrawComplexMode == true)
            {
                // Turn off the track mode
                winformsMap1.TrackOverlay.TrackMode = TrackMode.None;

                // First, find the center point of the created object
                Feature trackFeature = winformsMap1.TrackOverlay.TrackShapeLayer.InternalFeatures[0];
                
                // Set the drawing flag back to false
                inDrawComplexMode = false;
            }

        }

        private void winformsMap1_MouseUp(object sender, MouseEventArgs e)
        {
            // Exit if we are just clicking on the map.
            if (inDrawMode == true)
            {
                // Turn off the track mode
                winformsMap1.TrackOverlay.TrackMode = TrackMode.None;

                // Set the drawing flag back to false
                inDrawMode = false;

                return;
            }
        }
    }
}
